# Copyright (c) 2012-2016, 2018 Farsight Security, Inc.
# All Rights Reserved
#
# THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF FARSIGHT SECURITY, INC.
# The copyright notice above does not evidence any
# actual or intended publication of such source code.

import binascii
from collections import OrderedDict
import itertools
import json
import time
import traceback
import uuid
import urllib
import urlparse

import dnstable
import wdns

from dnstli.log import log, NOTICE
from dnstli.util import iso8601now
import dnstli.abort
import dnstli.conf
import dnstli.filter
import dnstli.fmt
import dnstli.util

conf = dnstli.conf.get_config()

class HTTPError(Exception):
    def __init__(self, status, message):
        self.status = status
        self.message = message + '\n'
        self.headers = [('Content-Type', 'text/plain')]

HTTPBadRequest = HTTPError('400 Bad Request', 'Error: unable to parse request.')
HTTPNotFound = HTTPError('404 Not Found', 'Error: no results found for query.')

class DNSTLI:
    def __init__(self, client):
        self.client = client

    def __call__(self, env, start_response):
        return Lookup(self.client, env, start_response)

class Lookup:
    def __init__(self, client, env, start_response):
        self.client = client
        self.env = env
        self.start_response = start_response
        self.delay_headers = True
        self.init_formatter()

        self.count = 0
        self.bytes = 0
        self.secs = 0
        self.secs_to_first = 0

    def init_formatter(self):
        if 'HTTP_ACCEPT' in self.env:
            ctypes = self.env['HTTP_ACCEPT'].split(',')
            ctypes = set([ x.strip().lower().split(';', 1)[0] for x in ctypes ])
            if 'application/json-seq' in ctypes:
                self.fmt = dnstli.fmt.JSONSeqFormatter()
                self.delay_headers = False
            elif 'application/json' in ctypes:
                self.fmt = dnstli.fmt.JSONFormatter()
            elif 'text/html' in ctypes:
                self.fmt = dnstli.fmt.HTMLFormatter()
            else:
                self.fmt = dnstli.fmt.TextFormatter()
        else:
            self.fmt = dnstli.fmt.TextFormatter()

    def __iter__(self):
        try:
            method, args = dnstli.util.get_method_args(self.env)

            if method == 'lookup' and args:
                self.begin_lookup(args)
                self.start_response('200 OK', [('Content-Type', self.fmt.content_type)])
                for chunk in dnstli.util.pager(self.format_response()):
                    self.bytes += len(chunk)
                    yield chunk
                self.log_lookup()
                return
            elif method == 'lookup' and not args:
                status = '200 OK'
                message = 'ok\n'
                headers = [('Content-Type', 'text/plain')]

                self.log_lookup_error(status)
                self.start_response(status, headers)
                yield message
                return
            else:
                raise HTTPBadRequest
        except dnstli.abort.Abort:
            self.log_lookup(early_exit=True)
        except GeneratorExit:
            self.log_lookup(early_exit=True)
            raise
        except HTTPError as err:
            if err is HTTPNotFound:
                self.log_lookup()
            else:
                self.log_lookup_error(err.status)
            self.start_response(err.status, err.headers)
            yield err.message
        except:
            status = '500 Internal Server Error'
            headers = [('Content-Type', 'text/html')]

            tb = traceback.format_exc()
            logid = str(uuid.uuid4())
            self.log_lookup_error(status, tb, logid)
            self.start_response(status, headers)
            yield dnstli.util.format_http500(logid)

    def log_lookup_error(self, status, tb=None, logid=None):
        msg = OrderedDict()
        msg['timestamp'] = iso8601now()
        msg['logtype'] = 'lookup-error'
        msg['host'] = self.env.get('SERVER_NAME', 'unset_SERVER_NAME')
        msg['authtype'] = self.env.get('dnstli.authtype', 'unset_authtype')
        msg['authuser'] = self.env.get('dnstli.authuser', 'unset_authuser')
        msg['ip'] = self.env.get('REMOTE_ADDR', 'unset_REMOTE_ADDR')
        msg['status'] = status
        if logid:
            msg['logid'] = logid
        if tb:
            msg['traceback'] = json.dumps(tb) # escape embedded quotes etc
        log(NOTICE, json.dumps(msg))

    def log_lookup(self, early_exit=False):
        msg = OrderedDict()
        msg['timestamp'] = iso8601now()
        msg['logtype'] = 'lookup'
        msg['host'] = self.env.get('SERVER_NAME', 'unset_SERVER_NAME')
        msg['authtype'] = self.env.get('dnstli.authtype', 'unset_authtype')
        msg['authuser'] = self.env.get('dnstli.authuser', 'unset_authuser')
        msg['ip'] = self.env.get('REMOTE_ADDR', 'unset_REMOTE_ADDR')
        msg['count'] = self.count
        msg['bytes'] = self.bytes
        msg['time'] = round(self.secs, 3)
        msg['ttfb'] = round(self.secs_to_first, 3)

        if early_exit:
            msg['early_exit'] = True

        query = getattr(self, 'query', None)
        if query is not None:
            msg['lookup'] = query.log

        log(NOTICE, json.dumps(msg))

    def begin_lookup(self, args):
        qs = urlparse.parse_qs(self.env.get('QUERY_STRING',''))

        time_filter = dnstli.filter.TimeFilter()
        time_filter.apply_query_string(qs)
        time_filter.apply_policy(self.env.get('dnstli.policy', None))

        try:
            limit = int(qs['limit'][0])
            if limit > conf['res_limit_max']:
                limit = conf['res_limit_max']
        except ValueError:
            raise HTTPBadRequest
        except:
            limit = conf['res_limit_default']

        try:
            index, index_type, args = args.split('/', 2)
        except:
            raise HTTPBadRequest

        try:
            if index == 'rrset':
                self.query = rrsetQuery(index_type, args, self.fmt,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'])
            elif index == 'rdata':
                self.query = rdataQuery(index_type, args, self.fmt,
                        time_first_before=time_filter.time_first_before,
                        time_first_after=time_filter.time_first_after,
                        time_last_before=time_filter.time_last_before,
                        time_last_after=time_filter.time_last_after,
                        timeout=conf['query_exit_check_interval'])
            else:
                raise HTTPBadRequest

            self.count = 0
            self.t0 = time.time()
            res = self.client.lookup(self.query.qobj, limit)
            if self.delay_headers:
                try:
                    n = res.next()
                    self.secs_to_first = time.time() - self.t0
                    self.res = itertools.chain([n], res)
                except StopIteration:
                    raise HTTPNotFound
            else:
                self.res = res
        except dnstable.DnstableException:
            # Note that dnstable query timeouts do *not* make it here
            raise HTTPBadRequest

    def format_response(self):
        yield self.fmt.format_header()

        for entry in self.res:
            self.count += 1
            yield self.query.format_entry(entry)

        self.secs = time.time() - self.t0
        perf = 'Returned %s %s in %.2f seconds.' % (self.count, self.query.fmt_type, self.secs)
        yield self.fmt.format_footer(perf)

class rrsetQuery:
    fmt_type = 'RRsets'
    fmt_log_type = 'rrset'

    def __init__(self, index_type, args, fmt, time_first_before=None, time_first_after=None, time_last_before=None, time_last_after=None, timeout=None):
        self.format_entry = fmt.format_rrset
        self.log = OrderedDict()
        self.log['type'] = self.fmt_log_type

        name, rrtype, bailiwick = dnstli.util.padded_split(args, '/', 3)[:3]

        if name:
            name = urllib.unquote(name)
        if bailiwick:
            bailiwick = urllib.unquote(bailiwick)

        if index_type == 'name':
            self.log['name'] = name
        elif index_type == 'raw':
            try:
                name = wdns.domain_to_str(binascii.unhexlify(name))
            except TypeError:
                raise HTTPBadRequest
            except wdns.NameException:
                raise HTTPBadRequest
            self.log['raw'] = name
        else:
            raise HTTPBadRequest

        self.qobj = dnstable.query(dnstable.RRSET, name, rrtype, bailiwick,
                time_first_before=time_first_before,
                time_first_after=time_first_after,
                time_last_before=time_last_before,
                time_last_after=time_last_after, timeout=timeout)

        if rrtype:
            self.log['rrtype'] = rrtype.upper()
        if bailiwick:
            self.log['bailiwick'] = bailiwick

class rdataQuery:
    fmt_type = 'RRs'
    fmt_log_type = 'rdata'

    def __init__(self, index_type, args, fmt, time_first_before=None, time_first_after=None, time_last_before=None, time_last_after=None, timeout=None):
        self.format_entry = fmt.format_rdata
        self.log = OrderedDict()
        self.log['type'] = self.fmt_log_type

        try:
            rdata, rrtype = dnstli.util.padded_split(args, '/', 2)

            if rdata:
                rdata = urllib.unquote(rdata)

            if rrtype:
                # verify rrtype is defined; else throw
                wdns.str_to_rrtype(rrtype)
        except:
            raise HTTPBadRequest

        if index_type == 'name':
            self.qobj = dnstable.query(dnstable.RDATA_NAME, rdata, rrtype,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after, timeout=timeout)
            self.log['name'] = rdata
        elif index_type in ('ip', 'ip6'):
            rdata = rdata.strip()
            rdata = rdata.replace(',', '/')
            self.qobj = dnstable.query(dnstable.RDATA_IP, rdata,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after, timeout=timeout)
            self.log['ip'] = rdata
        elif index_type == 'raw':
            self.qobj = dnstable.query(dnstable.RDATA_RAW, rdata, rrtype,
                    time_first_before=time_first_before,
                    time_first_after=time_first_after,
                    time_last_before=time_last_before,
                    time_last_after=time_last_after, timeout=timeout)
            self.log['raw'] = rdata
        else:
            raise HTTPBadRequest

        if rrtype:
            self.log['rrtype'] = rrtype.upper()
